const logger = require("../../common.js").createLogger();
const utils = require("../../utils");
const stringify = require("json-stringify");

const NeutrinoExternalInterface = function(neutrino, gameKey, gamestate) {
  this.getId = function() {
    return gameKey;
  };
  this.awardScore = function(playerId, score, cb) {
    logger.trace({ key: gameKey }, "award score called");
    let scoreObj = null;
    if (typeof score === "object") scoreObj = score;
    else scoreObj = { score: score };
    neutrino.awardScore(gameKey, playerId, scoreObj.score, scoreObj.message);
    if (cb) cb();
  };
  this.endGame = function(endGameReason, cb) {
    logger.info(
        { key: gameKey },
        "External plugin called for ending the game "
    );
    neutrino.endGame(gameKey, endGameReason, cb);
  };
  this.gameObjectByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      objs[id] = utils.gameObjectToExtGameObject(gamestate.gameObjectById(id));
    }
    return objs;
  };
  this.changeOwner = function(playerId, ...ids) {
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      obj.owner = playerId;
    }
  };
  this.getObjectsByType = function(type) {
    let objs = [];
    Object.values(gamestate.gameObjects).forEach(o => {
      if (o.objType === type) {
        objs.push(utils.gameObjectToExtGameObject(o));
      }
    });
    return objs;
  };

  this.getNextTurnPlayerId = function() {
    let playerInfo = gamestate.getNextPlayer();
    if (playerInfo) return playerInfo.playerId;
  };

  this.getCurrentTurnPlayerId = function() {
    let playerInfo = gamestate.currentTurn();
    if (playerInfo) return playerInfo.playerId;
  };

  this.getCurrentTurnIndex = function() {
    return gamestate.getCurrentTurnIndex();
  };

  this.getAllObjIds = function() {
    return Object.keys(gamestate.gameObjects);
  };

  this.getOwner = function(id) {
    return gamestate.gameObjects[id] ? gamestate.gameObjects[id].owner : null;
  };

  this.setGameProp = function(key, value) {
    gamestate.setProp(key, value);
  };

  this.getGameProp = function(key) {
    return gamestate.getProp(key);
  };

  this.setPlayerProp = function(playerId, key, value) {
    let player = gamestate.getPlayerByPlayerId(playerId);
    if (!player) return;
    return player.setProp(key, value);
  };

  this.getPlayerProp = function(playerId, key) {
    let player = gamestate.getPlayerByPlayerId(playerId);
    if (!player) return;
    return player.getProp(key);
  };

  this.getPlayerScore = function(playerId) {
    let player = gamestate.getPlayerByPlayerId(playerId);
    if (!player) return;
    return player.score;
  };

  this.getAllPlayerIds = function() {
    let playerIds = [];
    for (let player of gamestate.players) playerIds.push(player.playerId);
    return playerIds;
  };

  /**
   * Initializes collision map from redis for collision detection
   * @param cb of form cb(err, result)
   */
  this.initCollisionMap = function(cb) {
    neutrino.initCollision(gamestate, cb);
  };

  /**
   * First persists all game objects and then
   * updates all game objects, that have position changed, position into
   * the collision map
   * @param cb
   */
  this.updateCollisionMap = function(cb) {
    neutrino.saveGameState(gamestate, (err, res) => {
      if (err && typeof cb === "function") return cb(err, null);
      neutrino.updateCollisionObjs(gamestate, cb);
    });
  };

  /**
   * returns booleans if objIdB is supplied, or else return an array of objects
   * collided with objIdA
   */
  this.collidedObjsWithObject = function(objIdA, objIdB, cb) {
    neutrino.collidedObjsWithObject(
        gamestate.getRedisKey(),
        objIdA,
        objIdB,
        cb
    );
  };

  /**
   * returns list of objs colliding with the boundary
   * @param boundary should be of format {minX: 0, minY: 1, maxX: 2, maxY:3}
   * @param cb callback to be called after the result be found, of format cb(err, result)
   */
  this.collidedObjsWithBoundary = function(boundary, cb) {
    neutrino.collidedObjsWithBoundary(gamestate.getRedisKey(), boundary, cb);
  };

  this.syncGameStateServVars = function() {
    logger.trace("Calling external index to send server vars");
    neutrino.sendGameStateServVars(gameKey);
  };

  this.syncPlayerServVars = function(playerId) {
    neutrino.sendPlayerServVars(gameKey, playerId);
  };

  this.syncGameObjServVars = function(...objIds) {
    neutrino.sendGameObjServVars(gameKey, ...objIds);
  };

  this.forceUpdateGameObjects = function(...objTransList) {
    if (logger.trace())
      logger.trace(
          { key: gameKey },
          "Sending force update for game objects",
          stringify(objTransList)
      );
    neutrino.forceUpdateGameObjects(gameKey, ...objTransList);
  };
  this.addTransient = function(key, val) {
    gamestate.addTransient(key, val);
  };
  this.removeTransient = function(key) {
    gamestate.removeTransient(key);
  };
  this.getTransient = function(key) {
    return gamestate.getTransient(key);
  };
  this.nextTurn = function(turnData) {
    neutrino.nextTurn(gameKey, turnData);
  };
  this.continueTurn = function(turnData) {
    neutrino.continueTurn(gameKey, turnData);
  };
  this.save = function(cb) {
    neutrino.saveGameState(gamestate, (err, res) => {
      if (err) return cb(err);
      cb(null);
    });
  };
  this.getCurrentTurnExpiry = function() {
    let durationProp = gamestate.getProp("turnExpiry");
    if (!durationProp) return -1;
    return parseInt(durationProp);
  };
  this.issueSavepoint = function(savepointName, cb) {
    let gameMeta = neutrino.registeredGames[gamestate.gameType].meta;
    neutrino.issueSavePoint(gameKey, savepointName, cb);
  };
  this.restoreFromSavepoint = function(savepointname, cb) {
    neutrino.restoreFromSavePoint(gameKey, savepointname, cb);
  };
  this.getState = function() {
    return gamestate.state;
  };
  this.spawnObjects = function(objectList, isImmediate, targetPlayerId) {
    neutrino.spawnObjects(gameKey, objectList, isImmediate, targetPlayerId);
  };
  this.getConnectedPlayers = function() {
    return neutrino.getConnectedPlayers(gameKey);
  };
  this.getDuration = function() {
    return gamestate.duration || -1;
  };
  this.getScheduledStartTime = function() {
    return gamestate.scheduledStartTime;
  };

  this.gameObjectPositionByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      objs[id] = { x: obj.position.x, y: obj.position.y, z: obj.position.z };
    }
    return objs;
  };

  this.gameObjectVelocityByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      objs[id] = { x: obj.velocity.x, y: obj.velocity.y, z: obj.velocity.z };
    }
    return objs;
  };

  this.gameObjectForceByIds = function(...ids) {
    let objs = {};
    for (let id of ids) {
      let obj = gamestate.gameObjectById(id);
      objs[id] = { x: obj.force.x, y: obj.force.y, z: obj.force.z };
    }
    return objs;
  };

  this.destroyGameObjects = function(isImmediate, ...ids) {
    neutrino.destroyGameObjects(gameKey, ids, isImmediate);
  };

  this.getGameMeta = function() {
    let gameMeta = neutrino.registeredGames[gamestate.gameType].meta;
    return utils.gameMetaToExtGameMeta(gameMeta);
  };

  this.forfeitGameForPlayer = function(playerId, reason) {
    neutrino.forfeitGame(gameKey, playerId, reason);
  };

  this.abandonGame = function(reason) {
    neutrino.abandonGame(gameKey, reason);
  };

  this.getGameStateTimeElapsed = function() {
    return gamestate.elapsedTimeMillis;
  };

  this.forceUpdateForceGameObjects = function(toSend, ...objForceList) {
    if (logger.trace())
      logger.trace(
          { key: gameKey },
          "Force updating force for gameObjects",
          stringify(objForceList)
      );
    if (toSend)
      return logger.error(
          { key: gameKey },
          "Force update of force to clients is not implemented, dropping it"
      );
    for (let obj of objForceList) {
      let gameObj = gamestate.gameObjects[obj.id];
      if (obj.force.x != null) gameObj.force.x = obj.force.x;
      if (obj.force.y != null) gameObj.force.y = obj.force.y;
      if (obj.force.z != null) gameObj.force.z = obj.force.z;
    }
  };

  this.forceUpdateSyncVars = function(toSend, ...objSyncVarsList) {
    if (logger.trace())
      logger.trace(
          { key: gameKey },
          "Force updating sync vars for gameObjects",
          stringify(objSyncVarsList)
      );
    if (toSend)
      return logger.error(
          { key: gameKey },
          "Force update of sync vars to clients is not implemented, dropping it"
      );
    for (let obj of objSyncVarsList) {
      let gameObj = gamestate.gameObjects[obj.id];
      for (let [key, value] of Object.entries(obj.syncVars))
        gameObj.setSyncVar(key, value);
    }
  };

  this.setGameResultMessage = function(message) {
    logger.trace({ key: gameKey }, "Setting gameRes %s", message);
    gamestate.setProp("_gameResMsg", message);
  };
  this.sendSlackAlert = function(message) {
    neutrino.sendSlackAlert(gameKey, message);
  };
  this.markGameObjectTypeTransient = function(...types) {
    gamestate.markGameObjectTypeTransient(types);
  };
  this.publishGameEvent = function(type, playerIds = [], msg, cb) {
    // only msg types above 1000 are allowed for external plugin events
    if (type < 1000)
      return cb(
          new Error("Only msg types above 1000 are allowed for game plugins"),
          null
      );
    neutrino.eBus.sendGameEvent(
        gameKey,
        type,
        playerIds.map(p => gamestate.getPlayerByPlayerId(p).userId),
        msg,
        cb
    );
  };

  this.publishCustomEvent = function(eventName, msg, cb) {
    neutrino.eBus.sendCustomEvent(eventName, gameKey, msg, cb);
  };
};

/* **** NOTE ****
Any allowed functions that should work even after game has ended, should go
into the below list, otherwise after game ended the external interface plugin
functions just returns with a warn log.
 */
const getAllowedGetters = function() {
  return [
    "getId",
    "gameObjectByIds",
    "getCurrentTurnPlayerId",
    "getCurrentTurnIndex",
    "getAllObjIds",
    "getGameProp",
    "getOwner",
    "getPlayerProp",
    "getPlayerScore",
    "getAllPlayerIds",
    "addTransient",
    "removeTransient",
    "getCurrentTurnExpiry",
    "getState",
    "getConnectedPlayers",
    "getDuration",
    "getScheduledStartTime",
    "getGameMeta",
    "getGameStateTimeElapsed",
    "publishGameEvent",
    "publishCustomEvent"
  ];
};

module.exports = { NeutrinoExternalInterface, getAllowedGetters };
